"""This module contains tools for extracing data from the message file.

"""
import os
import re

import pandas as pd
import numpy as np

EIG_HEADER_PATTERN = re.compile('^\\s*E I G E N V A L U E S  at time = [\\d+-\\.E]+\\s*$', flags=re.MULTILINE)
EIG_END_PATTERN = re.compile('^\\s*$', flags=re.MULTILINE)

OFFSET = 1

def get_modes(filename, output_type='dict', i_analysis=0, underdamped_only=True, sort_by_wn=True):
    """Gets the modes from an Adams Solver message file.
    
    Parameters
    ----------
    filename : str
        Name of Adams Solver message (.msg) file.
    output_type : str, optional
        If 'DataFrame', returns a pandas `:obj:pd.DataFrame` if 'dict', returns a `:obj:dict`. (Default is 'dict')
    i_analysis : int, (optional)
        If the Adams Solver message (.msg) file contains multiple eigensolutions, this specifies which one to get. (Default is 0 which takes the first analysis)

    Returns
    -------
    `:obj:pd.DataFrame` or `:obj:dict`
        Table of modes

    """
    temp_file = _write_temp_file(filename, i_analysis=i_analysis)

    data = pd.read_csv(temp_file, skiprows=1, delim_whitespace=True, index_col=0)
    os.remove(temp_file)

    data.fillna(0, inplace=True)    
    data.rename(columns={'Real(cycles/second)': 'Real', 'Imag.(cycles/second)': 'Imag'}, inplace=True)    

    data['Wn'] = np.sqrt(data['Real']**2 + data['Imag']**2)    
    data['zeta'] = -data['Real'] / data['Wn']    
    data['Wd'] = data['Wn'] * np.sqrt((1 - data['zeta']**2))
    
    if underdamped_only is True:
        # Drop the critically/overdamped modes
        data.drop(data[data['Imag']==0].index, inplace=True)

    if sort_by_wn is True:
        # Sort by Wn
        data.sort_values('Wn', inplace=True)
    
    data.reset_index(drop=False, inplace=True)
    data.rename({'Number': 'Aview Number'}, inplace=True)
    data.index.rename('Mode', inplace=True)

    if output_type.lower() == 'dict':
        data.reset_index(inplace=True)
        data = data.to_dict('list')        

    return data

def _write_temp_file(filename, i_analysis=0):

    temp_filename = os.path.splitext(filename)[0] + '.tmp'

    with open(filename, 'r') as fid, open(temp_filename, 'w') as fid_out:        

        start_line = None
        end_line = None
        counter = 0
        for i_line, line in enumerate(fid):

            if start_line is None and EIG_HEADER_PATTERN.match(line) and counter==i_analysis:
                # If we've found the eigenvalue table, 
                start_line = i_line
                fid_out.write(line.replace('+/-', '   '))
            
            elif start_line is None and EIG_HEADER_PATTERN.match(line) and counter<i_analysis:
                # If we found an eigenvalue table, but not the one we're looking for, increase the counter
                counter += 1
            
            elif start_line is not None and EIG_END_PATTERN.match(line):
                # If we've reached the first blank line since the table started, break from the loop
                end_line = i_line
                break
                
            elif start_line is not None:
                # If we're in the table, write the line
                fid_out.write(line.replace('+/-', '   '))
        
        # Raise exceptions if lines not found
        if start_line is None or end_line is None:          
            if counter > 0:
                raise ValueError(f'Only found {counter} analys' + 'is' if counter==1 else 'es' + f', but {i_analysis+1} expected!')
            else:
                raise Exception(f'No modal table found!')
    
    return temp_filename



import matplotlib.pyplot as plt
from matplotlib import cm  
import numpy as np
from scipy.signal.windows import hann, hamming, blackman, boxcar
from scipy.signal import find_peaks
from . import get_res_data

from thornpy.signal import fft_watefall as _fft_waterfall

def fft_watefall(res_file, res, comp, percent_overlap=50, n_fft=1024, t_min=None, t_max=None, input_res=None, input_comp=None, input_conversion_factor=60/360, input_unit='RPM', response_unit=None, response_conversion_factor=1, psd=False, z_scale='linear', order_lines=None, f_range=None, clean_sig=None, return_order_cuts=None, title=None, vmin=None, vmax=None):
    """Genenerates a waterfall plot from data in an Adams result or request file.
    
    Parameters
    ----------
    res_file : str
        File name of Adams result (.res) or request (.req) file.
    res : str
        Name of the result in the adams dataset
    comp : str
        Name of the result component in the Adams dataset
    percent_overlap : int, optional
        Percent overlap for the FFT waterfall, by default 50
    n_fft : int, optional
        Number of points used in each FFT, by default 1024
    window : str, optional
        Type of window used for each FFT, by default 'hanning'
    t_min : float, optional
        Start time if cropping data, by default None
    t_max : float, optional
        End time if cropping data, by default None
    input_res : str, optional
        Name of the result to use as the input speed.  If none is given the x axis of the waterfall plot will be time, by default None
    input_comp : str, optional
        Name of the result component to use as the input speed., by default None
    input_conversion_factor : float, optional
        Conversion factor applied to input signal, by default 60/360 (deg/s to RPM)
    response_unit : str, optional
        Text to display on the response signal y axis, by default None
    input_unit : str, optional
        Text to display on the input signal y axis, by default None
    conversion_factor : float, optional
        Conversion factor applied to response signal, by default 1
    psd : bool, optional
        If True FFT will output in PSD. If False FFT will output Magnitude , by default False
    z_scale : str, optional
        The scaling of the values in the spec. 'linear' is no scaling. 'dB' returns the values in dB scale. `psd` is True, this is dB power (10 * log10). Otherwise this is dB amplitude (20 * log10). by default 'linear'
    orders : list, optional
        Adds order fans to the waterfall plots at each order in the list.  Only used if `input_comp` is given.
    f_range : list, optional
        [mininum frequency, maximum frequency].  Only affects plot limits
    clean_sig : float, optional
        If given, removes data points that exceed `clean_sig` multiplied by the standard deviation of the signal.
    
    Returns
    -------
    Figure
        Waterfall plot
        
    """    
    time, sig, _ = get_res_data(res_file, res, comp, t_min=t_min, t_max=t_max)

    if input_res is not None:
        _, input_sig, _ = get_res_data(res_file, input_res, input_comp, t_min=t_min, t_max=t_max)
    else:
        input_sig = None
        
    waterfall = _fft_waterfall(time, sig, percent_overlap=percent_overlap, n_fft=n_fft, title=title, t_min=t_min, t_max=t_max, input_sig=input_sig, input_conversion_factor=input_conversion_factor, input_unit=input_unit, response_unit=response_unit, response_conversion_factor=response_conversion_factor, psd=psd, z_scale=z_scale, order_lines=order_lines, f_range=f_range, clean_sig=clean_sig, return_order_cuts=return_order_cuts, vmin=vmin, vmax=vmax)

    return waterfall



"""Functions for manipulating adams files and data
"""
import os
import re
import subprocess

def get_simdur_from_msg(msg_file):
	"""Reads an Adams message file (.msg) and returns the total duration of the simulation.
	
	Parameters
	----------
	msg_file : str
		Filepath to an Adams message file (.msg)
	
	Raises
	------
	RuntimeError
		Returned if no simulation end time was found in the specified message file
	
	Returns
	-------
	float
		Total duration of the simulation

	"""
	found = False
	with open(msg_file, 'r') as fid:
		for line in fid:
			if re.match(' *command: sim(ulate)?/dyn(anmic)?.*, *end *=.*', line.lower()):
				duration = float(re.split('end *=',line.lower().replace(' ',''))[-1].split(',')[0])
				found=True
	
	# Raise an error if no duration found
	if not found:
		raise RuntimeError('No simulation end time was found in the specified message file!')
		
	return duration

def get_simdur_from_acf(acf_file):
	"""Reads an Adams command file (.acf) and returns the total duration of the simulation.
	
	Parameters
	----------
	acf_file : str
		Filepath to an Adams message file (.msg)
	
	Raises
	------
	RuntimeError
		Returned if no simulation end time was found in the specified message file
	
	Returns
	-------
	float
		Total duration of the simulation

	"""
	found = False
	with open(acf_file, 'r') as fid:
		for line in fid:
			if re.match('sim(ulate)?/dyn(anmic)?.*, *end *=.*', line.lower()):
				duration = float(re.split('end *=',line.lower().replace(' ',''))[-1].split(',')[0])
				found=True
	
	# Raise an error if no duration found
	if not found:
		raise RuntimeError('No simulation end time was found in the specified acf file!')
		
	return duration

def get_n_threads(adm_file):
	"""Searches `adm_file` for the NTHREADS statement and returns its value.
	
	Parameters
	----------
	adm_file : str
		Path to an Adams Dataset (.adm) file
	
	Returns
	-------
	int
		Number of threads set `adm_file`

	"""
	n_threads = 1
	with open(adm_file, 'r') as fid:		
		for line in fid:
				
			# If at the NTHREADS statement, rewrite it
			if re.match('^,[ \\t]*nthreads[ \\t]*=[ \\t]*\\d$', line, flags=re.I):				
				n_threads = int(line.split('=')[1].strip())
	
	return n_threads		

def set_n_threads(adm_file, n_threads):
	"""Changes or creates the NTHREADS option on the PREFERENCES statement in `adm_file`.
	
	Parameters
	----------
	adm_file : str
		File path to an Adams Dataset (.adm) file
	n_threads : int
		Number of threads to use when running the model specified in `adm_file`

	"""
	found = False
	with open(adm_file, 'r') as fid_old, open(adm_file + '.tmp', 'w') as fid_new:
		for line in fid_old:
				
			# If at the NTHREADS statement, rewrite it
			if re.match('^,[ \\t]*nthreads[ \\t]*=[ \\t]*\\d$', line, flags=re.I):
				fid_new.write(f', NTHREADS = {n_threads}\n')
				found = True

			# If the end is reached and the NTHREADS statement isn't found, create it
			elif re.match('^end[ \\t]*$', line, re.I) and not found:
				fid_new.write(f'PREFERENCES/\n, NTHREADS = {n_threads}\n!\n')
				fid_new.write(line)
			
			# If at a normal line, write it
			else:
				fid_new.write(line)
		
	# Delete the old adm file and replace with modified
	os.remove(adm_file)
	os.rename(adm_file + '.tmp', adm_file)

def solve(acf_file, wait=False):
	"""Runs Adams Solver to solve the model specified in `acf_file`
	
	Parameters
	----------
	acf_file : str
		Path to an Adams Command (.acf) File 

	"""	
	file = os.path.split(acf_file)[-1]
	command = '"{}" ru-s "{}"'.format(os.environ['ADAMS_LAUNCH_COMMAND'], file)
	startupinfo = subprocess.STARTUPINFO()
	startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
	cwd = os.path.dirname(acf_file) if os.path.dirname(acf_file) != '' else os.getcwd()
	proc = subprocess.Popen(command, cwd=cwd, startupinfo=startupinfo)

	if wait:
		proc.wait()
	
	return proc

class AdmFileError(Exception):
	pass
